#include <gint/display.h>

#define abs(x) ((x)>=0 ? (x) : (-x))

void dellipse(int xm, int ym, int a, int b, int c) {
  long x = -a, y = 0; /* II. quadrant from bottom left to top right */
  long e2 = b, dx = (1 + 2 * x) * e2 * e2; /* error increment */
  long dy = x * x, err = dx + dy;          /* error of 1.step */

  long a2 = (long)a * a;
  long b2 = (long)b * b;

  do {
    dpixel(xm - x, ym + y, c); /* I. Quadrant */
    dpixel(xm + x, ym + y, c); /* II. Quadrant */
    dpixel(xm + x, ym - y, c); /* III. Quadrant */
    dpixel(xm - x, ym - y, c); /* IV. Quadrant */
    e2 = 2 * err;
    if (e2 >= dx) {
      x++;
      err += dx += 2 * b2;
    } /* x step */
    if (e2 <= dy) {
      y++;
      err += dy += 2 * a2;
    } /* y step */
  } while (x <= 0);
  while (y++ < b) {        /* to early stop for flat ellipses with a=1, */
    dpixel(xm, ym + y, c); /* -> finish tip of ellipse */
    dpixel(xm, ym - y, c);
  }
}

void dellipserect(int x0, int y0, int x1, int y1,
                  int c) { /* rectangular parameter enclosing the ellipse */
  long a = abs(x1 - x0), b = abs(y1 - y0), b1 = b & 1; /* diameter */
  double dx = 4 * (1.0 - a) * b * b,
         dy = 4 * (b1 + 1) * a * a;      /* error increment */
  double err = dx + dy + b1 * a * a, e2; /* error of 1.step */
  if (x0 > x1) {
    x0 = x1;
    x1 += a;
  } /* if called with swapped points */
  if (y0 > y1)
    y0 = y1; /* .. exchange them */
  y0 += (b + 1) / 2;
  y1 = y0-b1; /* starting pixel */
  a = 8 * a * a;
  b1 = 8 * b * b;
  do {
    dpixel(x1, y0, c); /* I. Quadrant */
    dpixel(x0, y0, c); /* II. Quadrant */
    dpixel(x0, y1, c); /* III. Quadrant */
    dpixel(x1, y1, c); /* IV. Quadrant */
    e2 = 2 * err;
    if (e2 <= dy) {
      y0++;
      y1--;
      err += dy += a;
    } /* y step */
    if (e2 >= dx || 2 * err > dy) {
      x0++;
      x1--;
      err += dx += b1;
    } /* x */
  } while (x0 <= x1);
  while (y0 - y1 <= b) {   /* to early stop of flat ellipses a=1 */
    dpixel(x0 - 1, y0, c); /* -> finish tip of ellipse */
    dpixel(x1 + 1, y0++, c);
    dpixel(x0 - 1, y1, c);
    dpixel(x1 + 1, y1--, c);
  }
}

void dcircle(int xm, int ym, int r, int c)
{
    dellipse( xm, ym, r, r, c);
}


void dellipse_fill(int xm, int ym, int a, int b, int cback, int cborder) {
  long x = -a, y = 0; /* II. quadrant from bottom left to top right */
  long e2 = b, dx = (1 + 2 * x) * e2 * e2; /* error increment */
  long dy = x * x, err = dx + dy;          /* error of 1.step */

  long a2 = (long)a * a;
  long b2 = (long)b * b;

  do {
    dline(xm - x, ym + y, xm + x, ym + y, cback );
    dpixel(xm - x, ym + y, cborder); /* I. Quadrant */
    dpixel(xm + x, ym + y, cborder); /* II. Quadrant */

    dline(xm - x, ym - y, xm + x, ym - y, cback );
    dpixel(xm + x, ym - y, cborder); /* III. Quadrant */
    dpixel(xm - x, ym - y, cborder); /* IV. Quadrant */
    e2 = 2 * err;
    if (e2 >= dx) {
      x++;
      err += dx += 2 * b2;
    } /* x step */
    if (e2 <= dy) {
      y++;
      err += dy += 2 * a2;
    } /* y step */
  } while (x <= 0);
  while (y++ < b) {        /* to early stop for flat ellipses with a=1, */
    dpixel(xm, ym + y, cborder); /* -> finish tip of ellipse */
    dpixel(xm, ym - y, cborder);
  }
}

void dcircle_fill(int xm, int ym, int r, int cback, int cborder)
{
    dellipse_fill( xm, ym, r, r, cback, cborder );
}